#![feature(io)]

use std::default::Default;
use std::io::{stdin, Read};

#[derive(PartialEq, Hash, Clone)]
pub struct Coord(isize, isize);

mod position {

    pub struct Position {
        pub id: usize,
        pub coord: ::Coord
    }

    #[derive(Default)]
    pub struct State {
        pub objects: Vec<Position>
    }

    pub fn id_at(state: &::State, coord: &::Coord) -> Option<usize> {
        state.position.objects.iter()
            .find(|item| item.coord == *coord)
            .map(|item| item.id)
    }

    pub fn position_of(state: &::State, id: usize) -> Option<::Coord> {
        state.position.objects.iter()
            .find(|item| item.id == id)
            .map(|item| item.coord.clone())
    }

}

mod graphics {

    pub struct Sprite {
        pub id: usize,
        pub glyph: char
    }

    #[derive(Default)]
    pub struct State {
        pub objects: Vec<Sprite>
    }

    pub fn glyph_for(state: &::State, id: usize) -> char {
        state.graphics.objects.iter()
            .find(|item| item.id == id)
            .map(|item| item.glyph)
            .unwrap_or('?')
    }
}

#[derive(Default)]
pub struct State {
    position: position::State,
    graphics: graphics::State
}

fn mktest(state: &mut State, id: usize, coord: Coord, glyph: char) {
    state.position.objects.push(position::Position {id: id, coord: coord});
    state.graphics.objects.push(graphics::Sprite {id: id, glyph: glyph});
}

fn mkstate() -> State {
    let mut state: State = Default::default();

    mktest(&mut state, 42, Coord(1, 1), '@');
    mktest(&mut state, 57, Coord(8, 8), '$');

    state
}

fn cell_glyph(state: &State, coord: &Coord) -> char {
    match position::id_at(state, coord) {
        Some(id) => graphics::glyph_for(state, id),
        None => '.'
    }
}

fn key_dir(input: char) -> Coord {
    match input {
        'h' => Coord(-1, 0),
        'j' => Coord(0, 1),
        'k' => Coord(0, -1),
        'l' => Coord(1, 0),
        _ => Coord(0, 0)
    }
}

fn vadd(a: Coord, b: Coord) -> Coord {
    Coord(a.0 + b.0, a.1 + b.1)
}

fn find_player(state: &::State) -> Option<usize> {
    state.graphics.objects.iter()
        .find(|item| item.glyph == '@')
        .map(|item| item.id)
}

fn draw(state: &State) {
    clear_screen();
    for y in 0..10 {
        for x in 0..10 {
            let coord = Coord(x, y);
            let glyph = cell_glyph(state, &coord);
            print!("{}", glyph);
        }
        print!("\n");
    }
}

fn update(state: &mut State, input: char) {
    match find_player(state) {
        Some(player) => {
            let dir = key_dir(input);
            let old_position = position::position_of(state, player).unwrap();
            let new_position = vadd(old_position, dir);
            move_object(state, player, new_position);
        }
        None => {}
    }
}

fn move_object(state: &mut State, to_move: usize, new_position: Coord) {
    for object in state.position.objects.iter_mut() {
        if object.id == to_move {
            object.coord = new_position.clone();
        }
    }
}

fn main() {
    let mut state = mkstate();
    loop {
        match get_input() {
            Some(ch) => {
                update(&mut state, ch);
                draw(&state);
            },
            None => { break; }
        }
    }
}

fn clear_screen() {
    print!("\x1b[2J\n");
}

fn get_input() -> Option<char> {
    let mut x = stdin().chars();
    match x.next() {
        Some(v) => v.ok(),
        None => None
    }
}

